C
C  This file is part of MUMPS 5.8.2, released
C  on Mon Jan 12 15:17:08 UTC 2026
C
C
C  Copyright 1991-2026 CERFACS, CNRS, ENS Lyon, INP Toulouse, Inria,
C  Mumps Technologies, University of Bordeaux.
C
C  This version of MUMPS is provided to you free of charge. It is
C  released under the CeCILL-C license 
C  (see doc/CeCILL-C_V1-en.txt, doc/CeCILL-C_V1-fr.txt, and
C  https://cecill.info/licences/Licence_CeCILL-C_V1-en.html)
C
      MODULE CMUMPS_MPI3_MOD
      IMPLICIT NONE
      INTEGER, PARAMETER :: WIN_SYM_PERM272 = 272
      INTEGER, PARAMETER :: WIN_FILS273     = 273
      INTEGER, PARAMETER :: WIN_STEP274     = 274
      INTEGER, PARAMETER :: WIN_LRGROUPS275 = 275
      INTEGER, PARAMETER :: WIN_RG2L276     = 276
      END MODULE CMUMPS_MPI3_MOD
